# -*- coding: utf-8 -*-
"""Converts or add Thyrsis mesh to spatialite database

USAGE

   python -m thyrsis.database.mesh_to_spatialite dir dbname.sqlite

   create a mesh database from the input directory dir
"""
from builtins import str

from . import SRID, mesh_tables_exists, create_mesh_tables

from ..utilities import cleanup_dirname, read_site_param, Timer

from ..log import logger

from ..database import sqlite
from shapely.geometry import Polygon

import os
import re
import numpy

def mesh_to_spatialite(project_dir, dbname, project_SRID=None):
    """Create spatialite database from input project, this can be used to
    initialize a computation database

    :param project_dir: path of the project folder
    :type project_dir: string
    :param dbname: path of database to create
    :type dbname: string
    :param project_SRID: epsg code
    :type project_SRID: string

    :return: None
    """
    
    cleanup_dirname(project_dir)

    is_new_db = not os.path.exists(dbname)
    conn = sqlite.connect(dbname)
    #conn.execute("PRAGMA synchronous = OFF")
    #conn.execute("PRAGMA journal_mode = OFF")
    #conn.execute("PRAGMA locking_mode = EXCLUSIVE")
    #conn.execute("PRAGMA temp_store = MEMORY")
    #conn.execute("PRAGMA page_size = 65536")
    if is_new_db:
        cur = conn.cursor()
        cur.execute("SELECT InitSpatialMetaData(1)")

    cur = conn.cursor()
    if not mesh_tables_exists(cur):
        create_mesh_tables(cur, project_SRID=project_SRID)

    load_mesh(cur, os.path.join(project_dir, 'input'), project_SRID=project_SRID)
    load_contours(cur, os.path.join(project_dir, 'input'), os.path.join(project_dir, 'contours'), project_SRID=project_SRID)

    cur.execute("VACUUM ANALYSE")
    conn.commit()

def parse_file(file_name, nb_values_per_elem, nb_of_elem):
    """parse file that contains a list of space separated value
    returns a list of tuples with the line number in last position

    :param file_name: path of the project folder
    :type file_name: string
    :param nb_values_per_elem: number of value for a element, used to parse file
    :type nb_values_per_elem: integer
    :param nb_of_elem: number of element, used to parse file
    :type nb_of_elem: integer

    :return: list of tuples with the project data
    :rtype: list
    """
    logger.debug("convert " + file_name + "... ")
    data = []
    tup = []
    with open(file_name) as fil:
        for line in fil.readlines():
            if line[0] == '#': # remove comments
                continue
            for val in line.split():
                tup.append(float(val))
                if len(tup) == nb_values_per_elem:
                    data.append(tuple(tup+[len(data)+1]))
                    tup = []
    assert len(data) == nb_of_elem
    return data

def dat_files(directory):
    """List all .dat files in directory

    :param directory: path of a folder
    :type directory: string

    :return: list of .dat files
    :rtype: list
    """
    return [os.path.join(directory, f) \
            for f in os.listdir(directory) if re.match(r'.*\.dat', f) \
            and not f[0] == '.'] # ignore hidden files

def load_contours(cur, input_dir, contour_dir, project_SRID=None):
    """Load contours entries into the current database
    
    :param cur: cursor on  a sqlite connection
    :type cur: sqlite3.Cursor
    :param input_dir: path of a input folder, containing site.param
    :type input_dir: string
    :param contour_dir: path of a folder with contour file
    :type contour_dir: string
    :param project_SRID: epsg code
    :type project_SRID: string

    :return: None
    """
    if project_SRID:
        SRID = project_SRID
    values = []
    offset = read_site_param(os.path.join(input_dir, 'site.param'))['offset']
    for contour_file in dat_files(contour_dir):
        logger.notice("convert", contour_file)
        nom = os.path.basename(contour_file)[:-4]
        group = 'riviere' if nom not in ['contour', 'site'] else nom
        nom = 'domaine' if nom == 'contour' else nom
        with open(contour_file) as fil:
            # the first line contains column names,
            # last is a repeat of first point
            points = []
            for line in fil.readlines()[1:]:
                spl = line.split()
                points.append(str(float(spl[0])+offset[0]) + 
                    " " + str(float(spl[1])+offset[1]) +
                    " " + spl[2] )
            linestring = 'LINESTRING Z(' + ','.join(points) + ')'
            values.append((linestring, int(SRID), nom, group))
    cur.executemany("INSERT INTO contours (GEOMETRY, nom, groupe) "
            "VALUES (Simplify(GeomFromText(?, ?), 0.1), ?, ?)", values)

def load_mesh(cur, input_dir, project_SRID=None):
    """Load contours entries into the current database
    
    :param cur: cursor on  a sqlite connection
    :type cur: sqlite3.Cursor
    :param input_dir: path of a input folder, containing site.param
    :type input_dir: string
    :param project_SRID: epsg code
    :type project_SRID: string

    :return: None
    """
    """Create nodes table"""
    cleanup_dirname(input_dir)

    # read the ofsets from site.param
    site_param = read_site_param(os.path.join(input_dir, 'site.param'))

    #@todo check that no parameter other than ofset and infiltration_mm_per_year
    #      are needed

    node_file = os.path.join(input_dir, 'noeuds.sat')
    logger.debug("convert " + node_file)
    data = []
    timer = Timer()
    with open(node_file) as fil:
        for line in fil.readlines():
            spl = line.split()
            data.append((float(spl[0])+site_param['offset'][0], 
                float(spl[1])+site_param['offset'][1]))
    logger.debug("    "+timer.reset("parsing")+" for "+str(len(data))+" lines")
    if project_SRID : 
        cur.executemany("INSERT INTO noeuds (GEOMETRY) "
            "VALUES ( MakePoint( ?, ?, {}) )".format(project_SRID), data)
    else :
        cur.executemany("INSERT INTO noeuds (GEOMETRY) "
            "VALUES ( MakePoint( ?, ?, {}) )".format(SRID), data)
    logger.debug("    "+timer.reset("insertion"))

    nb_of_nodes = len(data)
    nodes = data

    cur.executemany("UPDATE noeuds SET potentiel = ? WHERE OGC_FID = ? ",
            parse_file(os.path.join(input_dir, 'potentiel.sat'),
                1, nb_of_nodes))


    # here we hack parse_file, because it's supposed to read one
    # element per line, instead we read potentie, grad_x and grad_y
    # as element with nb_of_nodes values per element
    data = parse_file(os.path.join(input_dir, 'htsoec.sat'),
                nb_of_nodes, 3)[:2]
    potentiel = numpy.array(data[0][:-1])
    # potentiel already read in potentiel.sat
    # we check consistency
    cur.execute("SELECT potentiel FROM noeuds ORDER BY OGC_FID")
    potentiel_ref = numpy.array([v for [v] in cur.fetchall()])
    assert numpy.max(numpy.abs(potentiel_ref - potentiel)) < 1e-3
    for i, col in enumerate(['debit_massique_eau']):
        values = [(v, j+1) for j, v in enumerate(data[i+1][:-1])]
        cur.executemany("UPDATE noeuds SET "+col+" = ? WHERE OGC_FID = ?", values)

    logger.debug("convert " + os.path.join(input_dir, 'gsm.sat'))
    data = []
    timer.reset()
    with open(os.path.join(input_dir, 'gsm.sat')) as gsm_sat:
        for line in gsm_sat:
            for nid in line[1:].split():
                data.append((int(nid),))
        logger.debug("    "+timer.reset("parsing")+" for "+str(len(data))+" lines")
        cur.executemany("INSERT INTO noeuds_contour( nid ) "
            "VALUES ( ? )", data)
        logger.debug("    "+timer.reset("insertion"))

    logger.debug("convert " + os.path.join(input_dir, 'ehim.sat'))
    data = []
    with open(os.path.join(input_dir, 'ehim.sat')) as ehim_sat:
        for line in ehim_sat:
            line = line.rstrip()
            if re.match(r'^%', line) or line == "ehim 4" or line == "n":
                continue
            spl = line.split()
            data.append((int(spl[0]), float(spl[1])))
        cur.executemany("INSERT INTO potentiel_impose( nid, valeur ) "
            "VALUES ( ?, ? )", data)

    cur.executemany("UPDATE noeuds SET epaisseur_zs = ? "
            "WHERE OGC_FID = ?",
            parse_file(os.path.join(input_dir, 'epaisseur.sat'),
                1, nb_of_nodes))

    cur.executemany("UPDATE noeuds SET epaisseur_zns = ? "
            "WHERE OGC_FID = ? ",
            parse_file(os.path.join(input_dir, 'epaisseurZNS.sat'),
                1, nb_of_nodes))

    cur.executemany("UPDATE noeuds SET epaisseur_aquifere = ? "
            "WHERE OGC_FID = ? ",
            parse_file(os.path.join(input_dir, 'epaquifere.sat'),
                1, nb_of_nodes))

    cur.executemany("UPDATE noeuds SET altitude = ? WHERE OGC_FID = ? ",
            parse_file(os.path.join(input_dir, 'mnt.sat'),
                1, nb_of_nodes))

    if os.path.exists(os.path.join(input_dir, 'infiltration.sat')):
        cur.executemany("UPDATE noeuds SET infiltration = - ? WHERE OGC_FID = ? ",
                parse_file(os.path.join(input_dir, 'infiltration.sat'),
                    1, nb_of_nodes))
    else:
        m_per_mm = 1e-3
        sec_per_year = 365.25*24*60*60
        infiltration_m_per_s = float(site_param['infiltration_mm_per_year']*m_per_mm)\
                /sec_per_year
        cur.execute("UPDATE noeuds SET infiltration = "+str(infiltration_m_per_s))

    cur.executemany("UPDATE noeuds SET altitude_mur = ? WHERE OGC_FID = ? ",
            parse_file(os.path.join(input_dir, 'mur.sat'),
                1, nb_of_nodes))

    cur.executemany("UPDATE noeuds SET permeabilite_x = ?, "
        "permeabilite_y = ? WHERE OGC_FID = ? ", \
            parse_file(os.path.join(input_dir, 'permnoeuds.sat'),
                2, nb_of_nodes))

    if os.path.exists(os.path.join(input_dir, 'permnoeuds.insat')):
        cur.executemany("UPDATE noeuds SET permeabilite_zns = ? "
            "WHERE OGC_FID = ? ", \
                parse_file(os.path.join(input_dir, 'permnoeuds.insat'),
                    1, nb_of_nodes))

    # useless if we recompute that when we create elements
    # for the moment the trigger doesn't work
    cur.executemany("UPDATE noeuds SET surface = ? WHERE OGC_FID = ? ",
            parse_file(os.path.join(input_dir, 'surfnoeuds.sat'),
                1, nb_of_nodes))

    cur.executemany("UPDATE noeuds SET tranmissivite = ? WHERE OGC_FID = ? ",
            parse_file(os.path.join(input_dir, 'transnoeuds.sat'),
                1, nb_of_nodes))

    cur.execute("UPDATE geometry_columns_statistics set last_verified = 0")
    cur.execute("SELECT UpdateLayerStatistics('noeuds')")


    cur.execute("DROP INDEX IF EXISTS mailles_a_idx")
    cur.execute("DROP INDEX IF EXISTS mailles_b_idx")
    cur.execute("DROP INDEX IF EXISTS mailles_c_idx")
    cur.execute("DROP INDEX IF EXISTS mailles_d_idx")

 
    elem_file = os.path.join(input_dir, 'mailles.sat')
    logger.debug("convert " + elem_file)
    data = []
    with open(elem_file) as fil:
        for line in fil.readlines():
            spl = line.split()
            data.append(tuple([int(v) for v in spl[1:]]
                +([None] if len(spl) == 4 else [])
                + [Polygon([nodes[int(n)-1] for n in (spl[1:] if len(spl) == 4 else spl[1:3]+spl[6:2:-1])]).wkt] ))

    if project_SRID :
        cur.executemany("INSERT INTO mailles (a, b, c, d, GEOMETRY) "
            "VALUES (?, ?, ?, ? , GeomFromText(?, {}))".format(project_SRID), data)
    else :
        cur.executemany("INSERT INTO mailles (a, b, c, d, GEOMETRY) "
            "VALUES (?, ?, ?, ? , GeomFromText(?, {}))".format(SRID), data)

    nb_of_elem = len(data)

    cur.executemany("UPDATE mailles SET permeabilite_x = ?, "
        "permeabilite_y = ? WHERE OGC_FID = ? ",
            parse_file(os.path.join(input_dir, 'permeabilites.sat'),
                2, nb_of_elem))

    data = parse_file(os.path.join(input_dir, 'darcy.sat'),
                5, nb_of_elem)
    v = numpy.array(data)
    cur.executemany("UPDATE mailles SET v_x = ?, v_y = ?, v_norme = ? WHERE OGC_FID = ? ", v[:,2:])

    cur.execute("CREATE INDEX mailles_a_idx ON mailles(a)")
    cur.execute("CREATE INDEX mailles_b_idx ON mailles(b)")
    cur.execute("CREATE INDEX mailles_c_idx ON mailles(c)")
    cur.execute("CREATE INDEX mailles_d_idx ON mailles(d)")

    cur.execute("UPDATE geometry_columns_statistics set last_verified = 0")
    cur.execute("SELECT UpdateLayerStatistics('mailles')")

# run as script if invoqued as such
if __name__ == "__main__":
    import sys
    if len(sys.argv) == 2 and sys.argv[1] in ["-h","--help"]:
        help(sys.modules[__name__])
        exit(0)

    logger.enable_console(True)
    logger.set_level('debug')

    if len(sys.argv) != 3:
        logger.error("wrong number of arguments (try mesh_to_spatialite.py --help)")
        exit(1)

    if os.path.exists(sys.argv[2]):
        os.remove(sys.argv[2])

    #FIX SRID
    project_SRID=None
    mesh_to_spatialite(sys.argv[1], sys.argv[2],project_SRID=project_SRID)

